package jp.muziyoshiz.wmi.web;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import jp.muziyoshiz.wmi.web.jdo.Information;
import jp.muziyoshiz.wmi.web.jdo.Knower;
import jp.muziyoshiz.wmi.web.jdo.KnowerIdCounter;
import jp.muziyoshiz.wmi.web.jdo.PMF;

import com.google.appengine.api.users.User;
import com.google.appengine.api.users.UserService;
import com.google.appengine.api.users.UserServiceFactory;

@SuppressWarnings("serial")
public class KnowerServlet extends HttpServlet {
    public void doGet(HttpServletRequest req, HttpServletResponse resp) throws IOException {
        // テスト用に何のページなのか表示
        resp.getWriter().println("<p>Knower</p>");

        // ユーザ認証必須にする
        UserService userService = UserServiceFactory.getUserService();
        String thisURL = req.getRequestURI();
        if (req.getUserPrincipal() == null) {
            resp.getWriter().println(
                    "<p>Please <a href=\"" + userService.createLoginURL(thisURL)
                            + "\">sign in</a>.</p>");
            return;
        } else {
            resp.getWriter().println(
                    "<p><a href=\"" + userService.createLogoutURL(thisURL) +
                    "\">logout</a></p>");
        }

        // ユーザIDその他を取得
        User currentUser = userService.getCurrentUser();

        // テスト用のユーザIDを表示
        resp.getWriter().println("<p>UserId: " + currentUser.getUserId() + "</p>");
        
        // 全取得
        Collection<Knower> results = getKnowerByUser(currentUser);
            
        PrintWriter writer = resp.getWriter();
        writer.println("<table border=\"1\">");
        writer.println("<tr><th>Key</th><th>ID</th><th>Name</th><th>CreatedAt</th><th>Information</th></tr>");
        
        for (Knower knower : results) {
            writer.println("<tr>");
            writer.println("<td>" + knower.getKey() + "</td>");
            writer.println("<td>" + knower.getId() + "</td>");
            writer.println("<td>" + knower.getName() + "</td>");
            writer.println("<td>" + knower.getCreatedAt() + "</td>");
            writer.println("<td>");
            if (knower.getInformationKeys() != null) {
                for (String key : knower.getInformationKeys()) {
                    Information info = getInformationByKey(key);
                    if (info == null) {
                        writer.println("NULL, ");
                    } else {
                        writer.println(info.getName());
                        writer.println(", ");
                    }
                }
            }
            writer.println("</td>");
            writer.println("</tr>");
        }
        
        writer.println("</table>");
        
        resp.getWriter().println("<a href=\"/knower.jsp\">Add knower</a>");
        resp.getWriter().println("<a href=\"/\">Back to index</a>");
    }
    
    public void doPost(HttpServletRequest req, HttpServletResponse resp) throws IOException {
        // ユーザ認証必須にする
        UserService userService = UserServiceFactory.getUserService();
        if (req.getUserPrincipal() == null) {
            String thisURL = req.getRequestURI();
            resp.getWriter().println(
                    "<p>Please <a href=\"" + userService.createLoginURL(thisURL)
                            + "\">sign in</a>.</p>");
            return;
        }

        // ユーザIDその他を取得
        User currentUser = userService.getCurrentUser();

        String name = null;
        List<String> informationKeys = new ArrayList<String>();

        if (req.getParameter("name") != null) {
            name = req.getParameter("name");
        }
        if (req.getParameterValues("informationKeys") != null) {
            for (String infoKey : req.getParameterValues("informationKeys")) {
                informationKeys.add(infoKey);
            }
        }
        
        // idを割り当て
        Long id = getNewKnowerIdOfUser(currentUser);
        
        // knowerを作成
        Knower knower = new Knower();
        knower.setKey(currentUser.getUserId() + "/" + id);
        knower.setId(id);
        knower.setName(name);
        knower.setCreatedAt(new Date(System.currentTimeMillis()));
        knower.setInformationKeys(informationKeys);

        PersistenceManager pm = PMF.createManager();
        try {
            pm.makePersistent(knower);
        } finally {
            pm.close();
        }

        resp.sendRedirect("/knower");
    }

    private Information getInformationByKey(String key) {
        // 検索
        PersistenceManager pm = PMF.createManager();
        try {
            Information info = pm.getObjectById(Information.class, key);

            // detached状態に変更
            return pm.detachCopy(info);

        } catch (JDOObjectNotFoundException e) {
            return null;
        } finally {
            pm.close();
        }
    }
    
    @SuppressWarnings("unchecked")
    private Collection<Knower> getKnowerByUser(User user) {
        // 全取得
        PersistenceManager pm = PMF.createManager();

        try {
            // 検索対象はKnower
            Query query = pm.newQuery(Knower.class);
            
            // キーの前方一致
            query.setFilter("key >= :keyStart && key < :keyEnd");
            
            try {
                // 検索のプレフィックス
                String prefix = user.getUserId() + "/";
                
                Collection<Knower> results =
                    (Collection<Knower>)query.execute(
                            prefix, prefix + '\uFFFD');
                
                return pm.detachCopyAll(results);
            } finally {
                query.closeAll();
            }
        } finally {
            pm.close();
        }
    }
    
    private Long getNewKnowerIdOfUser(User user) {
        PersistenceManager pm = PMF.createManager();
        try {
            pm.currentTransaction().begin();
            try {
                KnowerIdCounter counter = null;
                try {
                    counter = pm.getObjectById(KnowerIdCounter.class, user.getUserId());
                } catch (JDOObjectNotFoundException e) {
                    // 見つからなかった場合は新たに定義
                    counter = new KnowerIdCounter(user.getUserId(), 0L);
                }
                
                counter.increment();
                pm.makePersistent(counter);
                pm.currentTransaction().commit();
                
                return counter.getCounter();
                
            } finally {
                if (pm.currentTransaction().isActive()) {
                    pm.currentTransaction().rollback();
                }
            }
        } finally {
            pm.close();
        }
    }
}
